#include "board.h"
#include "player.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



const int pvalues[] = {
  0, 100, 500, 300, 325, 900, INF
};


void domove(boardt * board, move m);

void copyboard(const boardt * board, boardt * nboard) {
  memcpy(nboard, board, sizeof (boardt));
}

void init(boardt * board){
  int i, x, j;

  /* start empty */
  for(x = 0; x < BSIZE * BSIZE; x ++) board->b[x] = EMPTY;
 
  for (i = 0; i < BSIZE; i++){
    PIECEAT(board->b,i,1) = MAKE(WHITE,PAWN);
    PIECEAT(board->b,i,6) = MAKE(BLACK,PAWN);
  }

  /*white pieces*/
  PIECEAT(board->b,0,0) = MAKE(WHITE,ROOK);
  PIECEAT(board->b,1,0) = MAKE(WHITE,KNIGHT);
  PIECEAT(board->b,2,0) = MAKE(WHITE,BISHOP);
  PIECEAT(board->b,3,0) = MAKE(WHITE,QUEEN);
  PIECEAT(board->b,4,0) = MAKE(WHITE,KING);
  PIECEAT(board->b,5,0) = MAKE(WHITE,BISHOP);
  PIECEAT(board->b,6,0) = MAKE(WHITE,KNIGHT);
  PIECEAT(board->b,7,0) = MAKE(WHITE,ROOK);
 
  /*black pieces*/
  PIECEAT(board->b,0,7) = MAKE(BLACK,ROOK);
  PIECEAT(board->b,1,7) = MAKE(BLACK,KNIGHT);
  PIECEAT(board->b,2,7) = MAKE(BLACK,BISHOP);
  PIECEAT(board->b,3,7) = MAKE(BLACK,QUEEN);
  PIECEAT(board->b,4,7) = MAKE(BLACK,KING);
  PIECEAT(board->b,5,7) = MAKE(BLACK,BISHOP);
  PIECEAT(board->b,6,7) = MAKE(BLACK,KNIGHT);
  PIECEAT(board->b,7,7) = MAKE(BLACK,ROOK);
  
 

  for(i = 2; i < BSIZE-2 ; i++)
    for(j = 0; j < BSIZE ; j++)
      PIECEAT(board->b,j,i) = EMPTY;
 

  board->nextturn = WHITE;
  board->flags = 0xFF;
  board->wenpassant = -1;
  board->benpassant = -1;
  board->whitemb = 0;
  board->blackmb = 0; 
  board->bcastled = board->wcastled = false;
}

void execute_move(boardt * board, move m, int simulation) {

  int turn = board->nextturn;

  if((m.srcx == 4) && (m.srcy == 0) && (m.destx == 2) && (m.desty == 0) 
     && ((RANKAT(board->b, m.srcx, m.srcy)) == KING)) {
    PIECEAT(board->b,3,0) = PIECEAT(board->b,0,0);
    PIECEAT(board->b,0,0) = EMPTY;
    board->wcastled = true;
  } 
  else if ((m.srcx == 4) && (m.srcy == 7) && (m.destx == 6) && (m.desty == 7) 
	   && ((RANKAT(board->b, m.srcx, m.srcy)) == KING)) {
    PIECEAT(board->b,5,7) = PIECEAT(board->b,7,7);
    PIECEAT(board->b,7,7) = EMPTY;
    board->bcastled = true;
  }
  else if ((m.srcx == 4) && (m.srcy == 7) && (m.destx == 2) && (m.desty == 7) 
	   && ((RANKAT(board->b, m.srcx, m.srcy)) == KING)) {
    PIECEAT(board->b,3,7)=PIECEAT(board->b,0,7);
    PIECEAT(board->b,0,7)=EMPTY;
    board->bcastled = true;
  }
  else if ((m.srcx == 4) && (m.srcy == 0) && (m.destx == 6) && (m.desty == 0) 
	   && ((RANKAT(board->b, m.srcx, m.srcy)) == KING)) {
    PIECEAT(board->b,5,0)=PIECEAT(board->b,7,0);
    PIECEAT(board->b,7,0)=EMPTY;
    board->wcastled = true;
  }

       
  domove(board, m);

  if((RANKAT(board->b, m.destx, m.desty) == PAWN) && 
     ((m.desty == (BSIZE - 1) && turn == WHITE) || (m.desty == 0 && turn == BLACK)))
    PIECEAT(board->b, m.destx, m.desty) = MAKE(turn, m.promo);
  board->nextturn = NEGATE(board->nextturn);
}

void domove(boardt * board, move m) {
  /* handling en passant */
  int dy = m.desty-m.srcy;
  int dx = m.destx - m.srcx;
  int turn = board->nextturn;

  if( (COLORAT(board->b, m.destx, m.desty) != EMPTY)){
    if (turn == WHITE) board->whitemb =  board->whitemb + pvalues[RANKAT(board->b, m.destx, m.desty)];
    else  board->blackmb = board->blackmb + pvalues[RANKAT(board->b, m.destx, m.desty)];
  }
  if(dy !=1 && RANKAT(board->b,m.srcx,m.srcy) == PAWN){
    if(turn == BLACK)
      board-> benpassant = m.srcx; 
    else board-> wenpassant = m.srcx;
    }
  
  else {
    if(turn == BLACK)
      board-> benpassant = 9;
    else board-> wenpassant = 9;
  }
  
  if(dx != 0 && RANKAT(board->b,m.srcx,m.srcy) == PAWN && PIECEAT(board->b,m.destx,m.desty) == EMPTY){
    PIECEAT(board->b,m.destx,m.srcy) = EMPTY;
    if (turn == WHITE) board->whitemb = board->whitemb + pvalues[PAWN];
    else  board->blackmb = board->blackmb + pvalues[PAWN];
  }

  PIECEAT(board->b, m.destx, m.desty) = PIECEAT(board->b, m.srcx, m.srcy);
  PIECEAT(board->b, m.srcx, m.srcy) = EMPTY;

  if(RANKAT(board->b,m.destx,m.desty) == KING) {
    if(board->nextturn == WHITE) board->flags &= ~(FL_WKCASTLE & FL_WQCASTLE);
    else board->flags &= ~(FL_BKCASTLE & FL_BQCASTLE);
  }
  if(RANKAT(board->b,m.destx,m.desty) == ROOK  && m.srcx == 0) {
    if(board->nextturn == WHITE) board->flags &= ~ FL_WQCASTLE;
    else board->flags &= ~ FL_BQCASTLE;
  }
  if(RANKAT(board->b,m.destx,m.desty) == ROOK  && m.srcx == BSIZE -1){
    if(board->nextturn == WHITE) board->flags &= ~ FL_WKCASTLE;
    else board->flags &= ~ FL_BKCASTLE;
  }

 
}

int validate_move(const boardt * board, move m) {
  boardt nboard;
 
  int turn = board->nextturn;
  if (m.srcx == m.destx && m.desty == m.srcy) return false;
  if (m.srcx < 0 || m.srcy < 0 || m.destx < 0 || m.desty < 0) return false;
  if (m.srcx >= BSIZE || m.srcy >= BSIZE || m.destx >= BSIZE || m.desty >= BSIZE) return false;

  /* not their piece */
  if (COLORAT(board->b,m.srcx,m.srcy) != turn) {

    if (noisy) printf("You have to move a piece that you own!\n");

    return false;
  }
  /* trying to move on top of their own piece? */
  if (COLORAT(board->b,m.destx,m.desty) == turn) {
    if (noisy)printf ("Uh... you can't capture your own piece!\n");
    return false;
  }
  if (! checky(board, m)) {
    return false;
  }
  copyboard(board,&nboard);
  domove(&nboard, m);  
  return !check(&nboard);
}

int check(const boardt * board) {
  int x;
  move m;
  int turn = board->nextturn;
  int oppt = NEGATE(board->nextturn);
  boardt nboard;
  copyboard(board, &nboard);
  nboard.nextturn = oppt;

  for (x = 0; x < 64; x++) if (nboard.b[x] == MAKE(turn, KING)) {
    m.destx = x % 8;
    m.desty = (x - m.destx) / 8;
    for (x = 0; x < 64; x++) {
      if (COLOR(nboard.b[x]) == oppt) {
        m.srcx = x % 8;
        m.srcy = (x - m.srcx) / 8;
        if (checky(&nboard, m)) {
  
          if (noisy) printf("This puts your king in check!
                            (%d,%d to %d,%d)\n", m.srcx, m.srcy,
                            m.destx, m.desty);
  
          return true;
        }
      }
    }

    return false;
  }
 
 if (noisy) printf ("Oh my!! The king is Gone!!\n");
  return false;
}

int attacked(const boardt * board, int x, int y) {
  int xx,skingx, skingy;
  move m;
  boardt nboard;
  int oppt = NEGATE(board->nextturn);

  m.destx = x;
  m.desty = y;
  copyboard(board, &nboard);
  nboard.nextturn = oppt;

  for (xx = 0; xx < 64; xx++) {
    if (COLOR(nboard.b[xx]) == oppt) {
      m.srcx = xx % 8;
      m.srcy = (xx - m.srcx) / 8;
      if (checky(&nboard, m)) {
	
	return true;
      }
    }
  }
  return false;
}
 
void pindent(int p) {
  while(p--)
    printf(" ");
}

void printboard(const boardt * board, int indent) {
  int i, j, color, tilecolor, rank;
  unsigned char piece;
  pindent(indent);
  printf("    a   b   c   d   e   f   g   h  \n"); 
  pindent(indent);
  printf("  +---+---+---+---+---+---+---+---+\n");
  for(j = BSIZE-1; j >= 0; j--){
    pindent(indent);
    printf("%d |", j + 1);
 

    for(i = 0; i < BSIZE; i++){
      piece = PIECEAT(board->b,i,j);
      rank = RANK(piece);
      color = COLOR(piece);
      tilecolor = TILECOLOR(board->b,i,j);

      printf(" %c ", ptoc(piece));

      printf("|");
    }
    printf(" %d\n", j + 1);
    pindent(indent);
    printf("  +---+---+---+---+---+---+---+---+\n");

  }
  pindent(indent);
  printf("    a   b   c   d   e   f   g   h\n");
}



int queen(const boardt * board, move m) {
  if (bishop(board, m) ||
      rook  (board, m)) return true;
  else return false;
}


int bishop(const boardt * board, move m) {
  int dx, dy, xsign, ysign, x, y;
  dx = m.destx - m.srcx;
  dy = m.desty - m.srcy;
  if (abs(dx) != abs(dy)) return false;
  xsign = (dx > 0)?1:-1;
  ysign = (dy > 0)?1:-1;
  for (x = m.srcx + xsign, y = m.srcy + ysign; x < 8 && x >= 0;) {
    if (x == m.destx && y == m.desty) return true;
    if (PIECEAT(board->b, x, y) != EMPTY) return false;

    x += xsign;
    y += ysign;
  }
  return false;
}

int rook (const boardt * board, move m) {
  int dx, dy, xsign, ysign, x, y;
  dx = m.destx - m.srcx;
  dy = m.desty - m.srcy;
  if (dx && dy) { return false; }

  xsign = (dx == 0)?0:((dx > 0)?1:-1);
  ysign = (dy == 0)?0:((dy > 0)?1:-1);
  for (x = m.srcx + xsign, y = m.srcy + ysign; x < 8 && x >= 0 && y < 8 && y >= 0;) {
    if (x == m.destx && y == m.desty) return true;
    if (PIECEAT(board->b,x,y) != EMPTY) return false;

    x += xsign;
    y += ysign;
  }
  return false;
}

int knight (const boardt * board, move m) {
  int dx, dy;
  dx = m.destx - m.srcx;
  dy = m.desty - m.srcy;

  if (((abs(dx) == 2) && (abs(dy) == 1)) || 
      ((abs(dx) == 1) && (abs(dy) == 2))) return true;

  return false;

}

int king (const boardt * board, move m) {

  int dx, dy;
  int flags = board -> flags;  
  int turn = board ->nextturn;
  dx = m.destx - m.srcx;
  dy = m.desty - m.srcy;
  
  if (turn==WHITE && (flags & FL_WQCASTLE))
    if ((m.srcx==4 && m.srcy == 0 && m.destx==2 && m.desty==0)
        && ((PIECEAT(board->b,1,0)==EMPTY) && (PIECEAT(board->b,2,0)==EMPTY) && (PIECEAT(board->b,3,0) == EMPTY))
        && !attacked(board,2,0) && !attacked(board,3,0) && !attacked(board,4,0)) return true;
  
  if (turn==BLACK && (flags & FL_BQCASTLE))
    if ((m.srcx==4 && m.srcy == BSIZE-1 && m.destx==2 && m.desty== BSIZE-1)
	 &&((PIECEAT(board->b,1,BSIZE-1)==EMPTY) && (PIECEAT(board->b,2,BSIZE-1)==EMPTY) && (PIECEAT(board->b,3,BSIZE-1) == EMPTY))
         && !attacked(board,2,BSIZE-1) && !attacked(board,3,BSIZE-1) && !attacked(board,4,BSIZE-1)) 
      return true;
  
  if (turn==BLACK && (flags & FL_BKCASTLE))
    if ((m.srcx==4 && m.srcy == BSIZE-1 && m.destx==6 && m.desty==BSIZE-1)
	&&
	((PIECEAT(board->b,5,7)==EMPTY) && (EMPTY==PIECEAT(board->b,6,7)))
	&& !attacked(board,5,BSIZE-1) && !attacked(board,6,BSIZE-1) && !check(board)) return true ;
  
  if (turn==WHITE && (flags & FL_WKCASTLE))
    if ((m.srcx==4 && m.srcy == 0 && m.destx==6 && m.desty==0)
	&& ((PIECEAT(board->b,5,0) == EMPTY) && (PIECEAT(board->b,6,0)==EMPTY))
	&& !attacked(board,5,0) && !attacked(board,6,0) && !check(board)) return true;
 
  if ((abs(dx) > 1) || (abs(dy) > 1)) return false;
  return true;
  
}


int pawn (const boardt * board, move m) {
  int oppt;
  int dx, dy;
  int turn = board->nextturn;
  oppt = NEGATE(turn);

  dx = m.destx - m.srcx;
  dy = m.desty - m.srcy;
  
  if(m.desty == 0 || m.desty == BSIZE -1) 
      if(m.promo != QUEEN &&  m.promo != KNIGHT  &&  
	 m.promo != BISHOP &&  m.promo != ROOK){
	  return false;
      }
  
      if (turn == BLACK) {      
      if ((dy == -1) && (dx == 0) && 
	  PIECEAT(board->b, m.destx, m.desty) == EMPTY) return true;
      if ((dy == -2) && (dx == 0) &&
	  PIECEAT(board->b, m.destx,5) == EMPTY && 
	  PIECEAT(board->b, m.destx, m.desty) == EMPTY &&
	  (m.srcy ==6))  return true;
    
      if ((dy == -1) && (abs(dx) == 1) && 
	  (COLORAT(board->b, m.destx, m.desty) == WHITE)) return true;
      
      if ((dy == -1) && (abs(dx) == 1)&&(PIECEAT(board->b, m.destx, m.desty) == EMPTY)
	  && m.destx == board->wenpassant 
	&& (PIECEAT(board->b,m.destx,m.desty+1) == MAKE(WHITE,PAWN)) 
	  && (m.desty == 2)) return true;
      
      
      return false;
      
  } else { 
      
      if ((dy == 1) && (dx == 0) && 
	  PIECEAT(board->b, m.destx, m.desty) == EMPTY) return true;
      if ((dy == 2) && (dx == 0) &&
	  PIECEAT(board->b, m.destx,2) == EMPTY && 
	  PIECEAT(board->b, m.destx, m.desty) == EMPTY &&
	  (m.srcy == 1)) return true;
      if ((dy == 1) && (abs(dx) == 1) && 
	  (COLORAT(board->b, m.destx, m.desty) == BLACK)) return true;
      
      if ((dy == 1) && (abs(dx) == 1) && 
	  (PIECEAT(board->b, m.destx, m.desty) == EMPTY)
	  && (COLORAT(board->b, m.destx, m.desty) == BLACK)
	  && m.destx == board->benpassant
	  && PIECEAT(board->b,m.destx,m.desty-1) == MAKE(BLACK,PAWN)
	  && m.desty == 5) return true;
      
      
      return false;

  }

  return false;
}


int checky(const boardt * board, move m) {

  int (*testfun) (const boardt *, move);
  
  switch(RANKAT(board->b, m.srcx, m.srcy)) {
  case QUEEN: testfun = queen; break;
  case BISHOP: testfun = bishop; break;
  case ROOK: testfun = rook; break;
  case KNIGHT: testfun = knight; break;
  case KING: testfun = king; break;
  case PAWN: testfun = pawn; break;
  default: return false;
  }
  return testfun(board, m);
}


int gameover(const boardt * board) {
  int gameover = true;
  int i, j, k;
  move m;
  char * player;
  /* if only kings, draw (checking for legal moves
     never happens). */

  for(i = 0; i < 64; i++) {
    if ((board->b[i] != EMPTY) && (RANK(board->b[i]) != KING)) {
      /* there are non-draw pieces remaining. See if there are any legal moves. */
      for(j = 0; j < 64; j++) {
	if (board->b[j] != EMPTY)
	  for (k = 0; k < 64; k++) {
	    m.srcx = j & 7;
	    m.srcy = j >> 3;
	    m.destx = k&7;
	    m.desty = k >> 3;
	    m.promo = QUEEN;
	    if (validate_move(board,m)) { 
	      gameover = false; 
	      goto gameo_out; 
	    }
	    m.promo = KNIGHT;
	    if (validate_move(board,m)) { 
	      gameover = false; 
	      goto gameo_out; 
	    }
	    m.promo = ROOK;
	    if (validate_move(board,m)) { 
	      gameover = false; 
	      goto gameo_out; 
	    }
	    m.promo = BISHOP;
	    if (validate_move(board,m)) { 
	      gameover = false; 
	      goto gameo_out; 
	    }
	  }
      }
      break; /* gameover stays true */
    }
  }
 gameo_out: ;

  player = "white\0black" + board->nextturn*6;

  if (gameover && check(board)) {
    //printf("Game Over: %s lost.\n", player);
    return board->nextturn;
  } else if (gameover) {
    //intf("Game Over: STALEMATE!!!\n");
    return STALEMATE;
  }

  return KINGOK;
}

const char * mtos(const boardt * board, const move * m) {
  static char ms[128];
  sprintf(ms, "%c%d%c%d", 
	  m->srcx + 'a', m->srcy + 1,
	  m->destx + 'a', m->desty + 1);
  return ms;
}

/* This is a bit more permissive than FEN, but accepts all valid
   FEN strings as well. */
/* rnbqkbnr/pp1ppppp/8/2p5/4P3/8/PPPP1PPP/RNBQKBNR w KQkq c6 0 2 */
int init_fen(boardt * board, const char * f) {
  int c, b;

  printf("Initializing from FEN position '%s'\n", f);

  /* parse board part */
  b = 0;
  for(c = 0; f[c] != ' '; c++) {
    if ((f[c] >= 'a' && f[c] <= 'z') ||
	(f[c] >= 'A' && f[c] <= 'Z')) {
      switch (f[c] | 32) {
      case 'r': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, ROOK); break;
      case 'n': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, KNIGHT); break;
      case 'b': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, BISHOP); break;
      case 'k': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, KING); break;
      case 'q': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, QUEEN); break;
      case 'p': board->b[b++] = MAKE((f[c] & 32)?BLACK:WHITE, PAWN); break;
      default: printf ("Bad character %c in FEN\n", f[c]); return 0;
      }
    } else if (f[c] >= '0' && f[c] <= '9') {
      int z = f[c] - '0';
      for (;z > 0 && b < BSIZE * BSIZE; z--) board->b[b++] = EMPTY;
    } else if (f[c] != '/') {
      printf("Bad character 0x%02X in FEN\n", f[c]); 
      return 0;
    }
    if (b > BSIZE * BSIZE) { printf("Board too long in FEN\n"); return 0; }
  }
  
  if (b != BSIZE * BSIZE) { printf("Board too short in FEN\n"); return 0; }

  /* f[c] is ' ' */
  c++;
  if (f[c] == 'w') board->nextturn = WHITE;
  else if (f[c] == 'b') board->nextturn = BLACK;
  else { printf("Bad turn in FEN\n"); return 0; }

  c++;

  /* castling flags */
  board->flags = 0;
  for(; f[c] != ' '; c++) {
    switch (f[c]) {
    case '-': break;
    case 'K': board->flags |= FL_WKCASTLE; break;
    case 'Q': board->flags |= FL_WQCASTLE; break;
    case 'k': board->flags |= FL_BKCASTLE; break;
    case 'q': board->flags |= FL_BQCASTLE; break;
    default: printf("Bad character in FEN\n"); return 0;
    }
  }

  /* f[c] is ' ' */
  c++;
  /* do en passant */
  if (f[c] != '-') {
    if ((f[c] >= 'a' && f[c] <= 'h')) {
      if (f[c+1] == '3')
	board->wenpassant = f[c] - 'a';
      else if (f[c+1] == '6')
	board->benpassant = f[c] - 'a';
      else { printf("Bad en passant row in FEN\n"); return 0; }
    } else { printf("Bad en passant col in FEN\n"); return 0; }
  }

  /* ignore the move counts; we don't use those rules */

  /* so, done ! */

  return 1;
}
