#ifndef __BOARD_H_
#define __BOARD_H_



/* flags for castling.
   each side has two flags, one for its queen side rook
   and one for its king side rook. Moving the king resets
   both castle flags.
*/
#define FL_WQCASTLE (1<<0)
#define FL_WKCASTLE (1<<1)
#define FL_BQCASTLE (1<<2)
#define FL_BKCASTLE (1<<3)



/*
6 most significant bits of the char: Rank of piece
least 2 significant bits: color
*/

enum piece {
  PAWN = 1,
  ROOK = 2,
  KNIGHT = 3,
  BISHOP = 4,
  QUEEN = 5,
  KING = 6,
  NUMPIECES
};

#define PIECENAMES " PRNBQK"

#define WHITE 0
#define BLACK 1
#define EMPTY 2

#define NEGATE(c) ((c)^1)

#define CHECKMATE 3
#define STALEMATE 4
#define KINGOK 5
#define CHECK 1
#define PROMO 1 << 1
#define CAPTURE 1 << 2
#define NONCAPTURE 1 << 3


/* Macros for puttin together color and rank, returning color or rank
   and finding the same for a certain position on a board. and for
   getting the piece at a certain position */

#define MAKE(C,R) ((R<<2) | (C)) 
#define COLOR(X) ((X) & 3)
#define RANK(X) ((X) >> 2)
#define RANKAT(B,X,Y) ((B)[((Y)<<3)+(X)] >> 2)
#define COLORAT(B,X,Y) ((B)[((Y)<<3)+(X)] & 3)
#define PIECEAT(B,X,Y) ((B)[((Y)<<3)+(X)])


#define TILECOLOR(B,X,Y) (((Y)^(X))&1)
#define BSIZE 8
#define true 1
#define false 0
#define noisy 0

typedef struct mv { 
  int srcx; 
  int srcy; 
  int destx; 
  int desty;
  int promo;
  unsigned char flag;
} move;

typedef struct boardt {
  unsigned char b[BSIZE*BSIZE];
  unsigned char flags;
  char wenpassant;
  char benpassant;
  int nextturn;
  int blackmb;
  int whitemb;
  int wcastled;
  int bcastled;
} boardt;


void init(boardt * board);
int validate_move(const boardt * board, move m); 
void execute_move(boardt * board, move m, int simulate);
int gameover(const boardt * board);
int check(const boardt * board);
void printboard(const boardt * board, int indent);
int bishop (const boardt * board, move m);
int queen (const boardt * board, move m);
int rook (const boardt * board, move m);
int king (const boardt * board, move m);
int knight (const boardt * board, move m);
int checky(const boardt * board, move m);
int attacked(const boardt * board, int x, int y);
void copyboard(const boardt * board, boardt * nboard);

/* initialize a board using standard FEN notation 
   returns nonzero on success
*/
int init_fen(boardt * board, const char * fen);

// const char * ptos(unsigned char p);
#define ptoc(p) (PIECENAMES[RANK(p)] | ((COLOR(p) & 1)<<5))
const char * mtos(const boardt * board, const move * m);

/* move to utilities file */
void pindent(int);

#endif
