#include "board.h"
#include "player.h"
#include "moves.h"
#include "evaluate.h"
#include <stdio.h>
#include <stdlib.h>

int eval(const boardt* board){
   int material = 0, castle = 0, result, i, j, turn = board->nextturn, blackpawns = 0, whitepawns = 0, wmat, bmat;
   int pawnformation = 0, bishops = 0, rookbonus, tropism = 0, badbishops = 0, development, pawnstructure, oppt = NEGATE(turn);
   int totalmat, val, matdiff;
   unsigned char piece;
   pstruct pawnstuff;
   
  if((result = gameover(board)) == turn)
    return -INF ;
  else if(result == NEGATE(turn))
    return INF;
  else if(result == STALEMATE)
    return 1000; 
  analyzepawnstructure(board, &pawnstuff);
  
  wmat =  board->whitemb;
  bmat =  board->blackmb;

  if(turn == BLACK){
    if(board->flags & FL_BKCASTLE) castle ++;
    if(board->flags & FL_BQCASTLE) castle ++;
    if(board->flags & FL_WKCASTLE) castle --;
    if(board->flags & FL_WQCASTLE) castle --;
  }
  else {
    if(board->flags & FL_WKCASTLE) castle ++;
    if(board->flags & FL_WQCASTLE) castle ++;
    if(board->flags & FL_BKCASTLE) castle --;
    if(board->flags & FL_BQCASTLE) castle --;
  }
  for(i = 0; i < BSIZE; i++)
    for(j = 0; j < BSIZE; j++){
      
      if(PIECEAT(board->b, i, j) == MAKE(turn, KING)){
	tropism -= evalkingtropism(board, i, j);
	continue;
      }
      if (PIECEAT(board->b, i, j) == MAKE(oppt, BISHOP)){
	bishops -= evalbadbishops(board, &pawnstuff, i, j);
	continue;
      }
      if(PIECEAT(board->b, i, j) == MAKE(BLACK, PAWN)){
	blackpawns++;
	continue;
      }
      if (PIECEAT(board->b, i, j) == MAKE(WHITE, PAWN)){
	whitepawns++;
	continue;
      }
      continue;
    }
  totalmat = 5850 - wmat - bmat;

  if(wmat == bmat)
    material = 0;
  else if( wmat < bmat){
    matdiff = bmat - wmat;
    val = MIN( 2400, matdiff) + (matdiff*(12000 - totalmat)*blackpawns)/(6400 * blackpawns +1);
    material = (turn == BLACK)? val : -val;
  }else if( wmat > bmat){
    matdiff = wmat - bmat;
    val = MIN( 2400, matdiff) + (matdiff*(12000 - totalmat)*whitepawns)/(6400 * whitepawns +1);
    material = (turn == WHITE)? val : -val;
  }
  
    
  

  rookbonus = - evalrookbonus(board, &pawnstuff);
  development = - evaldevelopment(board, &pawnstuff);
  pawnstructure = - evalpawnstructure(board, &pawnstuff);
  
  
  return material + (3* castle) + tropism +  development + 
    pawnstructure + badbishops + rookbonus;
}



int evalkingtropism( const boardt* board, int kingx, int kingy){
  int score = 0;
  int turn = board->nextturn;
  int opt = NEGATE(turn);
  int piecey = 0, piecex = 0,i;
  

  for( i = 0; i < 64; i++ ){
    piecey = i / 8;
    piecex = i % 8;
   
    if (board->b[i] == MAKE(opt,ROOK)){
      score -= ( MIN(abs( kingy - piecey ),
		     abs( kingx - piecex ) ) << 1 );
      continue;
    }
    else if (board->b[i] == MAKE(opt,KNIGHT)){
      score += 5 - abs( kingy - piecey ) - abs( kingx - piecex );
      continue;
    }
    else if (board->b[i] == MAKE(opt,QUEEN)){
      score -= MIN(abs( kingy - piecey ),
		   abs( kingx - piecex ) );
      continue;
    }    
  }
  
  
  return score;
}


int evalrookbonus( const boardt* board, pstruct * pawnstuff){

  int score = 0,turn = board->nextturn;
  int opt = NEGATE(turn), x, y;
  
  for( y = 1; y < BSIZE; y+=(BSIZE-2))
    for( x = 0; x < BSIZE; x++){
      // Find a rook
      if (RANKAT(board->b,x,y) == ROOK){
	// Is this rook on the seventh rank?
	
	if ((opt == WHITE) && ( y == 7))
	  score += 22;
	if ((opt == BLACK) && ( y == 1))
	  score += 22;
	
	// Is this rook on a semi- or completely open file?
	if ( pawnstuff->MaxPawnFileBins[ x ] == 0 )
	  {
	    if (  pawnstuff->MinPawnFileBins[ x ] == 0 )
	      score += 10;
	    else
	      score += 4;
	  }
	
	// Is this rook behind a passed pawn?
	if ((opt == WHITE) && (pawnstuff->MaxPassedPawns[x] < BSIZE*y + x ))
	  score += 25;
	if ((opt == BLACK) && (pawnstuff->MaxPassedPawns[x] >  BSIZE*y + x))
	  score += 25;
      }
    }
  return score;
}


int evaldevelopment(const boardt* board, pstruct* pawnstuff){
  int score = 0,turn = board->nextturn, square;
  int opt = NEGATE(turn), x, y, cnt; 
  
  if ( opt == WHITE ){
    // Has the machine advanced its center pawns?
    if ( board->b[11] == MAKE(WHITE,PAWN))
      score -= 15;
    if ( board->b[12] == MAKE(WHITE,PAWN))
      score -= 15;
    
    // Penalize bishops and knights on the back rank
    for(square = 0; square < 8; square++ ){
      if (((board->b[square]) == MAKE(WHITE,KNIGHT) ) ||
	  (board->b[square]) == MAKE(WHITE,BISHOP))  
	score -= 10;
    }
    
    // Penalize too-early queen movement
    if (board->b[3] != MAKE(WHITE,QUEEN)){
      // First, count friendly pieces on their original squares
      cnt = 0;
      if(board->b[0] == MAKE(WHITE,ROOK))
	cnt++;
      if (board->b[1] == MAKE(WHITE,KNIGHT))
	cnt++;
      if(board->b[2] == MAKE(WHITE,BISHOP))
          cnt++;
      if(board->b[4] == MAKE(WHITE,KING))
	cnt++;
      if (board->b[5] == MAKE(WHITE,BISHOP))
	cnt++;
      if (board->b[6] == MAKE(WHITE,KNIGHT))
	cnt++;
      if (board->b[7] == MAKE(WHITE,ROOK))
	cnt++;
        score -= ( cnt << 3 );
    }
    
    // And finally, incite castling when the enemy has a queen on the board
    // This is a slightly simpler version of a factor used by Cray Blitz
    if (board->b[59] != MAKE(BLACK, QUEEN))
      {
        // Being castled deserves a bonus
        if (board->wcastled)
          score += 10;
        // small penalty if you can still castle on both sides
        else if (board->flags & (FL_WKCASTLE | FL_WQCASTLE))
          score -= 24;
        // bigger penalty if you can only castle kingside
        else if (board->flags & FL_WKCASTLE)
          score -= 40;
        // bigger penalty if you can only castle queenside
        else if (board->flags & FL_WQCASTLE)
          score -= 80;
        // biggest penalty if you can't castle at all
        else
          score -= 120;
      }
  }
  else{
    // Has the machine advanced its center pawns?
    if ( board->b[51] == MAKE(BLACK,PAWN))
      score -= 15;
    if ( board->b[52] == MAKE(BLACK,PAWN))
      score -= 15;
    
    // Penalize bishops and knights on the back rank
    for(square = 56; square < 64; square++ ){
      if (((board->b[square]) == MAKE(BLACK,KNIGHT) ) ||
	  (board->b[square]) == MAKE(BLACK,BISHOP))  
	score -= 10;
    }
    
    // Penalize too-early queen movement
    if (board->b[3] == MAKE(BLACK,QUEEN)) {
      // First, count friendly pieces on their original squares
      cnt = 0;
      if(board->b[56] == MAKE(BLACK,ROOK))
	cnt++;
      if (board->b[57] == MAKE(BLACK,KNIGHT))
	cnt++;
      if(board->b[58] == MAKE(BLACK,BISHOP))
	cnt++;
      if(board->b[60] == MAKE(BLACK,KING))
	cnt++;
      if (board->b[61] == MAKE(BLACK,BISHOP))
	cnt++;
      if (board->b[62] == MAKE(BLACK,KNIGHT))
	cnt++;
      if (board->b[63] == MAKE(BLACK,ROOK))
	cnt++;
      score -= ( cnt << 3 );
    }
    
    // And finally, incite castling when the enemy has a queen on the board
    // This is a slightly simpler version of a factor used by Cray Blitz
    if (board->b[3] != MAKE(BLACK, QUEEN)){
      // Being castled deserves a bonus
      if (board->bcastled)
	score += 10;
      // small penalty if you can still castle on both sides
      else if (board->flags & (FL_BKCASTLE | FL_BQCASTLE))
	score -= 24;
      // bigger penalty if you can only castle kingside
      else if (board->flags & FL_BKCASTLE)
	score -= 40;
      // bigger penalty if you can only castle queenside
      else if (board->flags & FL_BQCASTLE)
	score -= 80;
      // biggest penalty if you can't castle at all
      else
	score -= 120;
    }
  }
  return score;
}


int evalbadbishops( const boardt* board, pstruct* pawnstuff, int x, int y){

  int score = 0;
  
  if ((x % 2) == (y % 2))
    score -= ( pawnstuff->MaxPawnColorBins[0] << 3 );
  else
    score -= (pawnstuff-> MaxPawnColorBins[1] << 3 );
  
  
  return score;
}


int evalpawnstructure( const boardt* board, pstruct * pawnstuff){
  int score = 0, bin, turn = board->nextturn,i ;
  int opt = NEGATE(turn);
  // First, look for doubled pawns

  for(bin = 0; bin < 8; bin++ )
    if ( pawnstuff->MaxPawnFileBins[ bin ] > 1 )
      score -= 8;
  

  if ( ( pawnstuff->MaxPawnFileBins[ 0 ] > 0 ) && ( pawnstuff->MaxPawnFileBins[ 1 ] == 0 ) )
    score -= 15;
  if ( ( pawnstuff->MaxPawnFileBins[ 7 ] > 0 ) && ( pawnstuff->MaxPawnFileBins[ 6 ] == 0 ) )
    score -= 15;
  for(bin = 1; bin < 7; bin++ ){
    if ( ( pawnstuff->MaxPawnFileBins[ bin ] > 0 ) && ( pawnstuff->MaxPawnFileBins[ bin - 1 ] == 0 )
	 && ( pawnstuff->MaxPawnFileBins[ bin + 1 ] == 0 ) )
      score -= 15;
  }
  
  if ( pawnstuff->MaxTotalPawns == 8 )
    score -= 10;
  
  // Penalize pawn rams, because they restrict movement
  score -= 8 * pawnstuff->PawnRams;
  
  if ( opt == WHITE ){
    if ( pawnstuff->MaxMostAdvanced[ 0 ] < MIN( pawnstuff->MinMostBackward[ 0 ], pawnstuff->MinMostBackward[ 1 ] ) )
      score += ( 8 - ( pawnstuff->MaxMostAdvanced[ 0 ] >> 3 ) ) *
	( 8 - ( pawnstuff->MaxMostAdvanced[ 0 ] >> 3 ) );
    if ( pawnstuff->MaxMostAdvanced[ 7 ] < MIN( pawnstuff->MinMostBackward[ 7 ], pawnstuff->MinMostBackward[ 6 ] ) )
      score += ( 8 - ( pawnstuff->MaxMostAdvanced[ 7 ] >> 3 ) ) *
	( 8 - ( pawnstuff->MaxMostAdvanced[ 7 ] >> 3 ) );
    for( i = 1; i < 7; i++ ){
      if ( ( pawnstuff->MaxMostAdvanced[ i ] < pawnstuff->MinMostBackward[ i ] ) &&
	   ( pawnstuff->MaxMostAdvanced[ i ] < pawnstuff->MinMostBackward[ i - 1 ] ) &&
	   ( pawnstuff->MaxMostAdvanced[ i ] < pawnstuff->MinMostBackward[ i + 1 ] ) )
	score += ( 8 - ( pawnstuff->MaxMostAdvanced[ i ] >> 3 ) ) *
	  ( 8 - ( pawnstuff->MaxMostAdvanced[ i ] >> 3 ) );
    }
  }
  else {
    if ( pawnstuff->MaxMostAdvanced[ 0 ] > MAX( pawnstuff->MinMostBackward[ 0 ], pawnstuff->MinMostBackward[ 1 ] ) )
      score += ( pawnstuff->MaxMostAdvanced[ 0 ] >> 3 ) *
	( pawnstuff->MaxMostAdvanced[ 0 ] >> 3 );
    if ( pawnstuff->MaxMostAdvanced[ 7 ] > MAX( pawnstuff->MinMostBackward[ 7 ], pawnstuff->MinMostBackward[ 6 ] ) )
      score += ( pawnstuff->MaxMostAdvanced[ 7 ] >> 3 ) *
	( pawnstuff->MaxMostAdvanced[ 7 ] >> 3 );
    for( i = 1; i < 7; i++ )
      {
        if ( ( pawnstuff->MaxMostAdvanced[ i ] > pawnstuff->MinMostBackward[ i ] ) &&
             ( pawnstuff->MaxMostAdvanced[ i ] > pawnstuff->MinMostBackward[ i - 1 ] ) &&
             ( pawnstuff->MaxMostAdvanced[ i ] > pawnstuff->MinMostBackward[ i + 1 ] ) )
          score += ( pawnstuff->MaxMostAdvanced[ i ] >> 3 ) *
	    ( pawnstuff->MaxMostAdvanced[ i ] >> 3 );
      }
  }
  
  return score;
}





void analyzepawnstructure( const boardt* board, pstruct* pawnstuff){
  int i, square, turn = board->nextturn;
  int opt = NEGATE(turn), x, y;
  
  for( i = 0; i < 8; i++ ){
    pawnstuff->MaxPawnFileBins[ i ] = 0;
    pawnstuff->MinPawnFileBins[ i ] = 0;
  }
  pawnstuff->MaxPawnColorBins[ 0 ] = 0;
  pawnstuff->MaxPawnColorBins[ 1 ] = 0;
  pawnstuff->PawnRams = 0;
  pawnstuff->MaxTotalPawns = 0;
  
  if ( opt == WHITE ){
    for( i = 0; i < 8; i++ ){
      pawnstuff-> MaxMostAdvanced[ i ] = 63;
      pawnstuff->MinMostBackward[ i ] = 63;
      pawnstuff->MaxPassedPawns[ i ] = 63;
    }
    for( square = 8; square <= 55; square++ ){
  
      if ( board->b[square]  == MAKE(WHITE,PAWN) ){
	y = square/8;
	x = square % 8;
	
	// This pawn is now the most advanced of all white pawns on its file
	pawnstuff->MaxPawnFileBins[x]++;
	pawnstuff->MaxTotalPawns++;
	pawnstuff->MaxMostAdvanced[y] = square;
	
	if ((y%2)==(x%2))
	  pawnstuff->MaxPawnColorBins[ 0 ]++;
	else
	  pawnstuff->MaxPawnColorBins[ 1 ]++;
	
	if ( board->b[square - 8] == MAKE(BLACK,PAWN))
	  pawnstuff->PawnRams++;
      }
      else if ( board->b[square] == MAKE(BLACK,PAWN)){
	x = square % 8;
	pawnstuff->MinPawnFileBins[x]++;
          pawnstuff->MinMostBackward[x] = square;
      }
    }
  }
  else{
    for(i = 0; i < 8; i++ ){
      pawnstuff-> MaxMostAdvanced[ i ] = 0;
      pawnstuff->MaxPassedPawns[ i ] = 0;
      pawnstuff->MinMostBackward[ i ] = 0;
    }
      for(square = 55; square >= 8; square-- ){
        if (board->b[square] == MAKE(BLACK,PAWN)){
	  y = square / 8;
          x = square % 8;
	  
	  pawnstuff->MaxPawnFileBins[x]++;
          pawnstuff->MaxTotalPawns++;
          pawnstuff->MaxMostAdvanced[x] = square;
	  
          if ((y % 2) == (x % 2))
            pawnstuff->MaxPawnColorBins[ 0 ]++;
          else
            pawnstuff->MaxPawnColorBins[ 1 ]++;
	  
          if (board->b[square - 8] == MAKE(WHITE,PAWN))
            pawnstuff->PawnRams++;
        }
        else if (board->b[square]  == MAKE(WHITE,PAWN)){
          x = square % 8;
          pawnstuff->MinPawnFileBins[x]++;
          pawnstuff->MinMostBackward[x] = square;
        }
      }
  }
  
}

