#ifndef __TEMPO_H_
#define __TEMPO_H_

#include <core.h>
#include <poplib.h>
#include <string.h>
#include <list.h>
#include <char.h>


open Core;
open String;
open List;


/* flags for castling.
   each side has two flags, one for its queen side rook
   and one for its king side rook. Moving the king resets
   both castle flags.
*/ 
#define FL_WQCASTLE (1<<0)
#define FL_WKCASTLE (1<<1)
#define FL_BQCASTLE (1<<2)
#define FL_BKCASTLE (1<<3)



/*
6 most significant bits of the char: Rank of piece
least 2 significant bits: color
*/


#define PAWN  1
#define ROOK  2
#define KNIGHT  3
#define BISHOP  4
#define QUEEN  5
#define KING  6


#define PIECENAMES " PRNBQK"

#define WHITE 0
#define BLACK 1
#define EMPTY 2

#define NEGATE(c) ((c)^1)

#define CHECKMATE 3
#define STALEMATE 4
#define KINGOK 5
#define CHECK 1
#define PROMO 1 << 1
#define CAPTURE 1 << 2
#define NONCAPTURE 1 << 3


/* Macros for puttin together color and rank, returning color or rank
   and finding the same for a certain position on a board. and for
   getting the piece at a certain position */

#define MAKE(C,R) ((R<<2) | (C)) 
#define COLOR(X) ((X) & 3)
#define RANK(X) ((X) >> 2)
#define RANKAT(B,X,Y) ((B)[((Y)<<3)+(X)] >> 2)
#define COLORAT(B,X,Y) ((B)[((Y)<<3)+(X)] & 3)
#define PIECEAT(B,X,Y) ((B)[((Y)<<3)+(X)])


#define TILECOLOR(B,X,Y) (((Y)^(X))&1)
#define BSIZE 8
#define noisy false

#define ptoc(p) (PIECENAMES[RANK(p)] | ((COLOR(p) & 1)<<5))
#define abs(x) ((x > 0) ? x : -x)




#define MINI -1
#define MAXI 1 
#define INF 0x7FFFFFF
#define MAXMOVES 300





struct move { 
  int srcx; 
  int srcy; 
  int destx; 
  int desty;
  char promo;
}

struct boardt {
  unsigned byte b[BSIZE*BSIZE];
  unsigned byte flags;
  byte wenpassant;
  byte benpassant;
  int nextturn;
  int blackmb;
  int whitemb;
  bool wcastled;
  bool bcastled;
}

extern struct pstruct {
  int MaxPawnFileBins[8];
  int MaxPawnColorBins[2];
  int MaxTotalPawns;
  int PawnRams;
  int MaxMostAdvanced[8];
  int MaxPassedPawns[8];
  int MinPawnFileBins[8];
  int MinMostBackward[8];
  
}

#define full_dep <*(cord_id, string)> list

#endif
