

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "board.h"
#include "player.h"

#define MAXLINE 200

enum {
  XBOARD = 4,
  NEW,
  FORCE,
  GO,
  MOVE,
  DRAW,
  HINT,
  BK,
  UNDO,
  REMOVE,
  IGNORE,
  PROTOVER,
  PING,
  EDIT,
  SETBOARD,
  OOPS,
};

void driver();

void usage ();
void usage () {
    printf("usage:\n   tempo [-debug] [-board \"FEN notation\"] [-interactive hh,hc,ch,cc]\n"
	   "(defaults to xboard mode)\n\n");
    exit(-1);
}

int main(int argc,char* argv[]){
 
  char* options, *input;
  move (*white) (const boardt *);
  move (*black) (const boardt *);
  int result;
  move nextmove;
  boardt board;
  init(&board);

  if(argc > 1 && !strcmp(argv[1], "-debug")) {
    debug ++;
    argv ++;
    argc --;
  }

  /* allow -board "fen notation" */
  if(argc > 1 && !strcmp(argv[1], "-board")) {
    if (argc <= 2 || !init_fen(&board, argv[2])) usage ();
    /* this is nasty. we need a new parameter parsing scheme. */
    argv += 2;
    argc -= 2;
  } 
  
  if(argc > 1 && !strcmp(argv[1], "-interactive")) {
    /* XXX move to "interactive" function */

    options = argv[2];
    if (argc < 3) usage ();

    switch(options[0]){
    case 'h':
      white = prompt_next_move;
      //whitepromo = prompt_promotion;
      break;
    case 'c':
      white = get_next_move;
      //whitepromo = get_promotion;
      break;
    default: usage(); return -1;
    }
    switch(options[1]){ 
    case 'h':
      black = prompt_next_move;
      //blackpromo = prompt_promotion;
      break;
    case 'c':
      black = get_next_move;
      //blackpromo = get_promotion;
      break;
    default: usage(); return -1;
    }
  
    printf("starting game:\n");

    while((result = gameover(&board))== KINGOK) {
      printf("printing board: (it's now the turn of player %d)\n", board.nextturn + 1);
      printboard(&board, 0);
      switch (board.nextturn){
      case WHITE: nextmove = white(&board); break;
      case BLACK: nextmove = black(&board); break;
      }
      printf("executing move:(%c,%d)->(%c,%d)\n",
	     nextmove.srcx+'a',
	     nextmove.srcy+1,
	     nextmove.destx+'a',
	     nextmove.desty+1);
      if (validate_move(&board, nextmove))
	execute_move(&board, nextmove, false);
      else printf ("Invalid move.\n");
    }

    printf("--------- Game Over --------\n");
    printboard(&board, 0);

    switch(result){
    case STALEMATE : printf("\nDraw! (Stalemate)\n\n");break;
    case WHITE : printf("\nBlack mates!\n\n");break;
    case BLACK : printf("\nWhite mates!\n\n");
    }

  } else driver ();
 
  return 0;
}

void driver(){
  boardt board;
  char input[MAXLINE], promotion, fen[MAXLINE];
  int token, started =  false;
  int srcx, srcy, destx, desty, promo, n;
  int myturn = EMPTY, result;
  char* pieces;
  move nextmove, tempmove;
  pieces = " prnbqk";  
  setbuf(stdout, NULL);

  //readline(input);
  fgets(input, MAXLINE, stdin);
  // fprintf(stderr,"%s\n",input);
  while(strncmp("quit",input,4) != 0){
    //fprintf(stderr,"%s\n",input);

    if (strncmp("xboard", input, 6) == 0)
      token = XBOARD; 
    else if (strncmp("protover", input, 8) == 0 )
      token = PROTOVER;
    else if (strncmp("go", input, 2) == 0)
      token = GO;
    else if (strncmp("new", input, 3) == 0)
      token = NEW;
    else if (strncmp("force", input, 5) == 0)
      token = FORCE;
    else if (strncmp("white", input, 5) == 0)
      token = WHITE;
    else if (strncmp("black", input, 5) == 0)
      token = BLACK;
    else if (strncmp("ping", input, 4) == 0){
      token = PING;
      sscanf(input, "ping %d", &n);
    } 
    else if (strncmp("draw", input, 4) == 0)
      token = DRAW;
    else if (strncmp("hint", input, 4) == 0)
      token = HINT; 
    else if (strncmp("bk", input, 2) == 0)
      token = BK;
    else if (strncmp("edit", input, 4) == 0)
      token = EDIT;
    else if (strncmp("accepted", input, 8) == 0)
      token = OOPS;
    else if (strncmp("rejected", input, 8) == 0)
      token = OOPS;
    else if (strncmp("undo", input, 4) == 0)
      token = UNDO;
    else if (strncmp("remove", input, 6) == 0)
      token = REMOVE;
    else if (strncmp("variant", input, 7) == 0)
      token = OOPS;
    else if (strncmp("random", input, 6) == 0)
      token = OOPS;
    else if (strncmp("level", input, 5) == 0)
      token = OOPS;
    else if (strncmp("time", input, 4) == 0)
      token = OOPS;
    else if (strncmp("otim", input, 4) == 0)
      token = OOPS;
    else if (strncmp("st", input, 2) == 0)
      token = OOPS;
    else if (strncmp("sd", input, 2) == 0)
      token = OOPS;
    else if (strncmp("?", input, 1) == 0)
      token = OOPS;
    else if (strncmp("result", input, 6) == 0)
      return;    
    else if (strncmp("otim", input, 4) == 0)
      token = OOPS;
    else if (strncmp("setboard", input, 8) == 0){
      token = SETBOARD;
      sscanf(input, "setboard %s", fen);
    }
    else if (strncmp("hard", input, 4) == 0)
      token = OOPS;
    else if (strncmp("name", input, 4) == 0)
      token = OOPS; 
    else if (strncmp("post", input, 4) == 0)
      token = OOPS; 
    else if (strncmp("nopost", input, 6) == 0)
      token = OOPS;
    else if (strncmp("analyze", input, 6) == 0)
      token = OOPS;
    else if (strncmp("name", input, 4) == 0)
      token = OOPS;
    else if (strncmp("rating", input, 6) == 0)
      token = OOPS; 
    else if (strncmp("ics", input, 3) == 0)
      token = OOPS;
    else if (strncmp("computer", input, 8) == 0)
      token = OOPS;
    else if (strncmp("pause", input, 5) == 0)
      token = OOPS;
    else if (strncmp("resume", input, 6) == 0)
      token = OOPS;
    else if (strlen(input) == 5){
      token = MOVE;
      nextmove.srcx = input[0] - 'a';
      nextmove.srcy = input[1] - '1';
      nextmove.destx = input[2] - 'a';
      nextmove.desty = input[3] - '1';
     }
    else if (strlen(input) == 6){
      token = MOVE;
      nextmove.srcx = input[0] - 'a';
      nextmove.srcy = input[1] - '1';
      nextmove.destx = input[2] - 'a';
      nextmove.desty = input[3] - '1';
      switch(input[4]){
      case 'r':
	nextmove.promo =  ROOK;
      case 'n':
	nextmove.promo = KNIGHT;
      case 'b':
	nextmove.promo = BISHOP;
      case 'q':
	nextmove.promo = QUEEN;
      default:
	printf("Illegal move: %s\n", input);
      }
    } else token = OOPS;
    /*
      #define XBOARD 4
      #define NEW 5
      #define FORCE 6
      #define GO 7
      #define MOVE 8
      #define DRAW 9
      #define HINT 10
      #define BK 11
      #define UNDO 12
      #define REMOVE 13
      #define IGNORE 14
      #define PROTOVER 15
      #define PING 16
      #define EDIT 17
    */
    switch(token){
    case OOPS: break;
    case XBOARD: break;
    case NEW: 
	started = true;
	init(&board); 
	myturn = BLACK;
	break;
    case FORCE:  
	myturn = EMPTY; break;
    case GO:  
	if(started){
	myturn = board.nextturn;
	nextmove = get_next_move(&board);
	execute_move(&board, nextmove, true); 
	if((nextmove.desty == 7 || nextmove.desty ==0) 
	   && RANKAT(board.b, nextmove.srcx, nextmove.srcy) == PAWN)
	  printf( "move %c%d%c%d%c\n",nextmove.srcx+'a',
		  nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1, pieces[nextmove.promo]);
	else 
	  printf( "move %c%d%c%d\n",nextmove.srcx+'a',
		  nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1); 
	} else printf("Error (can't go... you didn't start yet :P): go\n");
	break;
    case MOVE: 
	if (started && validate_move(&board, nextmove))
	    execute_move(&board, nextmove, true);
	else{ 
	    if((nextmove.desty == 7 || nextmove.desty ==0) 
	       && RANKAT(board.b, nextmove.srcx, nextmove.srcy) == PAWN)
		printf( "Illegal move %c%d%c%d%c\n",nextmove.srcx+'a',
			nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1, pieces[nextmove.promo]);
	      else 
		  printf( "Illegal move %c%d%c%d\n",nextmove.srcx+'a',
			  nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1);
	}
	if (myturn == board.nextturn){
	    nextmove = get_next_move(&board);
	    execute_move(&board, nextmove, true); 
	    if((nextmove.desty == 7 || nextmove.desty ==0) 
	       && RANKAT(board.b, nextmove.srcx, nextmove.srcy) == PAWN)
		printf( "move %c%d%c%d%c\n",nextmove.srcx+'a',
			nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1, pieces[nextmove.promo]);
	    else 
		  printf( "move %c%d%c%d\n",nextmove.srcx+'a',
			  nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1);
	}
    
	else printf("Error (can't go... you didn't start yet :P): go\n");
	break;
    case DRAW:
	break;
    case HINT: 
	tempmove = get_next_move(&board);
	if((nextmove.desty == 7 || nextmove.desty ==0) 
	   && RANKAT(board.b, nextmove.srcx, nextmove.srcy) == PAWN)
	    printf( "Hint: %c%d%c%d%c\n",nextmove.srcx+'a',
		    nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1, pieces[nextmove.promo]);
	else 
	    printf( "Hint: %c%d%c%d\n",nextmove.srcx+'a',
		    nextmove.srcy + 1, nextmove.destx + 'a', nextmove.desty + 1);
	break;
    case UNDO: printf("Error (can't undo): undo\n"); break; 
    case REMOVE: printf("Error (can't remove): remove\n"); break;  
    case BK: printf("I don't know what you're looking for :P\n"); break;
    case PROTOVER:  printf("feature setboard=1 ping=0 playother=0 san=0"
			   " usermove=0 time=0 draw=0 sigint=0 sigterm=1"
			   " reuse=1 analyze=0 colors=0 ics=0 name=0"
			   " pause=0 done=1\n"); break;
    case EDIT: printf("Error (can't edit): edit\n"); break; 
    case WHITE:
	board.nextturn = WHITE;
	myturn = BLACK; 
	break;
    case BLACK:
	board.nextturn = BLACK;
	myturn = WHITE; 
	break;
    case SETBOARD:
	init_fen(&board, fen);
    }     
    
    if(started){
	result = gameover(&board);
	switch (result){
	case STALEMATE: printf("1/2-1/2 {Stalemate}\n"); fflush(stdout); break;
	case WHITE: printf("0-1 {Black mates}\n"); fflush(stdout); break;
	case BLACK: printf("1-0 {White mates}\n"); fflush(stdout);
	}
    }
    
    fgets(input, MAXLINE, stdin);
    token = OOPS;
  }
  return ;
  
  
}
