#include "board.h"
#include "player.h"
#include "moves.h"
#include "evaluate.h"
#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h> 

/* might be set by command line flags */
int debug = DEBUG;

const int piecevalues[] = {
  0, 100, 500, 300, 325, 900, INF
};

move get_next_move(const boardt * board) {
  return search(board);
}

move prompt_next_move(const boardt * board) {
  move blah;
  char sc, dc, inp[10], p = 0;
  int sr, dr;
  printf("enter your move:\n");
  
  //readline(inp);
  fgets(inp, 10, stdin);
  if(strlen(inp) == 5)
    sscanf(inp, "%c%d%c%d", &sc, &sr, &dc, &dr);
  else if(strlen(inp) == 6)
     sscanf(inp, "%c%d%c%d%c", &sc, &sr, &dc, &dr, &p);
  blah.srcx = sc - 'a';
  blah.destx = dc - 'a';
  blah.srcy = sr - 1;
  blah.desty = dr - 1;
  switch(p){
  case 'r':
      blah.promo = ROOK; break;
  case 'n':
    blah.promo = KNIGHT; break;
  case 'b':
    blah.promo = BISHOP; break;
  case 'q':
    blah.promo = QUEEN; break;
  default:
    blah.promo = 0;
  }
  printf("promo: %c\n", p);
  return blah;
}
/*
int promote(const boardt * board, int x, int y, int simulation) {
  if(simulation)
    return get_promotion(board,x,y);
  else
    if(board->nextturn == WHITE) return whitepromo(board, x, y);
    else return blackpromo(board, x, y);
}

int prompt_promotion(const boardt * board, int x, int y) {
  char promo;
  printf("Congratulations! You are now  the proud owner of a promotable pawn.\n"
	 "What would you like to promote him/her to?\n"
	 "(enter a lowercase letter please)\n");
  scanf("\n%c", &promo);
  switch(promo){
  case 'r':
    return ROOK;
  case 'n':
    return KNIGHT;
  case 'b':
    return BISHOP;
  case 'q':
    return QUEEN;
  default:
    printf("Say what?\n");
    return prompt_promotion(board, x, y);
  }
  
}
*/
int get_promotion(const boardt * board, int x, int y) {
  return QUEEN;//TODO: fix this
}



/*
  evaluation function:
  (if you've won, -infty, if lost, infty)
  - material
  - spaces you threaten
     - weighted by location
  - spaces you occupy
     - weighted by location

  - mobility
    - in opening,
     - control of center of board
     - ability/execution of castling
     - pawn structure: avoid doubled and passed pawns

  WE WANT A LOW VALUE!!


*/

int evaluate(boardt* board){
  int material = 0, castle = 0, result, turn = board->nextturn;
 
  if((result = gameover(board)) == turn)
    return -INF ;
  else if(result == NEGATE(turn))
    return INF;
  else if(result == STALEMATE)
    return 1000; 
  
  material += (turn == WHITE)? board->whitemb : board->blackmb;
  material -= (turn == BLACK)? board->whitemb : board->blackmb;

  if(turn == BLACK){
    if(board->flags & FL_BKCASTLE) castle ++;
    if(board->flags & FL_BQCASTLE) castle ++;
    if(board->flags & FL_WKCASTLE) castle --;
    if(board->flags & FL_WQCASTLE) castle --;
  }
  else {
    if(board->flags & FL_WKCASTLE) castle ++;
    if(board->flags & FL_WQCASTLE) castle ++;
    if(board->flags & FL_BKCASTLE) castle --;
    if(board->flags & FL_BQCASTLE) castle --;
  }
  
  return material + (3* castle) ;
}

move search(const boardt * board){
 
  move bestmove, tempm;
  int worstval = -INF,  bestval = INF ,tempv1, tempv,tempv2,i;
  boardt nboard;
  move movearray[MAXMOVES];
  int number = sortedmoves(board, movearray);
  if(debug)
    printf("# of moves: %d\n", number);

  for(i = 0; i < number; i++){
    tempm = movearray[i];
    copyboard(board, &nboard);
    
    execute_move(&nboard,tempm, true);
    
    //tempv = evaluate(&nboard);
    //tempv = scout(&nboard, 5, MAXI, -INF-INF, bestval);
    tempv = scout2(&nboard, 4, -INF, bestval);
    //tempv = eval(&nboard);
    if (debug)
      printf("      (%s: %d)\n", mtos(board, &tempm), tempv);

    if(tempv < bestval){
      bestmove = tempm;
      bestval = tempv;
    }
    
    if(tempv > worstval){
    worstval = tempv;
    }
  }

  return bestmove;

  
}

int scout(boardt* board,int n, int minormax, int a, int b){
  int i = 0, low, high, x, val, number;
  move tempm;
  boardt nboard;
  move movesarray[MAXMOVES];
 
  
  /* Now, we'll end up picking a shortest move sequence to a checkmate, 
     since we weigh the depth at which we find the checkmate*/
  if(n == 0) return evaluate(board);
  if(gameover(board) != KINGOK) return n * evaluate(board);
  //if(gameover(board) != KINGOK|| (n == 0))  return evaluate(board) ;

  number = sortedmoves(board,  movesarray);
  low = a;
  high = b;
  
  if(minormax == MAXI) val = -(INF);
  else val = INF;

  while(i < number) {
    tempm = movesarray[i];
    i++;
    copyboard(board,&nboard);
 
    execute_move(&nboard, tempm, true);
    if(minormax == MAXI)
      x = scout(&nboard, n-1, MINI, low, high);
    else 	
      x = scout(&nboard, n-1, MAXI, low, high);
    
    if(minormax == MAXI){
      if((x > low) && (x < b) && (i > 1))
	x = scout(&nboard, n-1, MINI, low, b);
      low = (low > x) ? low : x;
      val = (val > x) ? val : x;
      if(val >= b) return val;
      high = low+1;
    } 
    else{
      if((x < high) && (x > a) && (i > 1))
	x = scout(&nboard, n-1, MAXI, a, x);
      high = (high < x) ? high : x;
      val = (val < x) ? val : x;
      if(val <= a) return val;
      low = high - 1;
    }
  }
  
  return val;
  
}


int scout2(boardt* board, int n, int alpha, int beta){
  int i, number, b, s;
  move tempm;
  boardt nboard;
  move movesarray[MAXMOVES];
 
  
  /* Now, we'll end up picking a shortest move sequence to a checkmate, 
     since we weigh the depth at which we find the checkmate*/
  if(n == 0) return eval(board);
  if(gameover(board) != KINGOK) return n * eval(board);

  number = sortedmoves(board,  movesarray);


   
  copyboard(board,&nboard);
  tempm = movesarray[0];
  execute_move(&nboard, tempm, true);
  /*  if (tempm.flag == CHECK  && n == 1) 
    b = - scout2(&nboard, 1 , -beta, -alpha);
    else*/ 
	b = - scout2(&nboard, n - 1, -beta, -alpha);
  if(b >= beta) return b;
  if(b > alpha) alpha = b;
  
  for(i = 1; i < number; i++){ //easily parallelizable
    tempm = movesarray[i];
    copyboard(board,&nboard);
    execute_move(&nboard,tempm,true);
    s = - scout2(&nboard, n-1, -alpha -1, -alpha);
    if (s > b ) b = s;
    if (s >= beta) return s;
    if (s > alpha) {
      /*      if (tempm.flag == CHECK  && n == 1)
	s = - scout2( &nboard, n, -beta, -alpha);
	else */
	s = - scout2( &nboard, n-1, -beta, -alpha);
      if (s >= beta) return s;
      if (s > alpha) alpha = s;
      if (s > b) b = s;
    }
      
  }

  return b;
}

int nummoves(const boardt* board){
  int i,j,k,l,m = 0;
  move mv;
  mv.promo = QUEEN;

  for(i = 0; i < BSIZE; i++)
    for(j = 0; j < BSIZE; j++)
      for(k = 0; k < BSIZE; k++)
	for(l = 0; l < BSIZE; l++){
	  mv.srcx = i;
	  mv.srcy = j;
	  mv.destx = k; 
	  mv.desty = l;

	  if(validate_move(board, mv)) m++;
	}
  return m;
}

