#ifndef __GRIDSIM_H_
#define __GRIDSIM_H_

#include <hashtable.h>
#include <core.h>
#include <string.h>
#include <list.h>

#define MAXCORDS 10000
#define cord_id int

open Hashtable;
open Core;
open String;
open List;
/*
datatype dep =   One of cord_id
| And of dep
| Or of dep list
*/

union dep { 
    cord_id One;
    <dep> list And;
    <dep> list Or;
}

/*
  datatype full_dep = fOne string
  | fAnd of string
  | fOr of string option list 
*/

union full_dep {
    string fOne;
    <<full_dep> Opt > list fAnd;
    <<full_dep> Opt > list fOr;
}
   


/*
  If a certain cord has a result.forward(f) then it is a forwarding address f
  otherwise, if it is a result.result(r) then r is the result string.
 */
union result { 
    string result;
    cord_id forward;
}


/*
  a cord that would be spawed on the grid would be composed of a function f,
  an argument arg that it should be run on, its dependencies deps and a result
  option res, wich is either "none" if the cord is yet to have been completely run,
  or "some" result.
 */
struct cord {
    result f (string, full_dep);
    string arg;
    dep deps;
    <result> Opt  res;
}


/*
  initializes all the global data structures needed for the simulation.
 */

extern void init_sim(); //CALL THIS TO USE THIS SIMULATION!!

/*
  a cord is to call this function before returning. 
  NOTE: in the real thing, a cord should be able to just exit without having to return. 
  But in this simulation, a cord must call exit_cord with its result and then return.
  the return value may be anything at all (it doesn't matter);
 */
extern void exit_cord (result result);

/*
  spawn(f,a,d) puts out a cord with dependecies d, function f and argument a, 
  to be run once the dependencies are filled. It returns the cord_id of that cord.
 */
extern cord_id spawn ( result f(string, full_dep), string arg, dep deps);


/*
  query(i) returns either null (in the case where the cord with id i has not finished yet)
  or "some" of the result. As an invariant, this function won't return a forwarding address.
  It chases all the forwards.
 */
extern <result>Opt query (cord_id);


#endif
